#ifndef _BTREE_H
#define _BTREE_H
/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* btree.h
   binary tree
*/

#include "nstring.h"

/* NODEDATA can be defined somewhere else */
#ifndef NODEDATA
#define NODEDATA /*nothing*/
#endif


typedef struct _NODE
{
   int nodesize;
   NSTRING name;
   struct _NODE * low;
   struct _NODE * high;
   NODEDATA
} NODE;


/* function that can be called by bt_foreach() : */
typedef void (* BT_PROC)( NODE * );


NODE * bt_mknode( int namlen, char * nam, int nodsize );
/* allocates a new node.
   the name is not copied ! just the pointer is kept
   nodsize: sizeof(NODE)
   */

NODE * bt_mktree( char * dummy, int nodsize );
/* make a new tree by allocating a node with a dummy name.
   dummy must be 0-terminated !
   nodsize: sizeof(NODE)
   */

NODE * bt_search( NODE * tree, int namlen, char * nam, int * comp );
/* searches nam in the tree
     if found     -> sets *comp  = 0 and returns a pointer to the node
     if not found -> sets *comp != 0 and returns a pointer to
                     the node where the new one must be inserted.
          Insert at node->low  if *comp < 0, else at node->high :
             newnode = bt_mknode( nam, len );
             if( *comp < 0 )
                node->low  = newnode;
             else
                node->high = newnode;
*/

NODE * bt_insert( NODE * root, int namlen, char * nam, int * isnew );
/* searches and inserts nam into the tree
   returns the node containing nam
   *isnew is set to 1 if nam was inserted, 0 if it was already there */

void bt_foreach( NODE * root, NODE * node, BT_PROC proc );
/* calls function 'proc' for each node of the tree
   depth first walking, i.e. keys are ordered
   the node "root" is ignored (where the dummy element is) */


#endif /* _BTREE_H */

