/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* conmsg.c
   see scan.h
*/

#include "scan.h"
#include "btree.h"
#include "salloc.h"


static char * kindstr[] =
{
   "WARNING",
   "ERROR",
   "FATAL ERROR",
   "DBG MSG"
};


static NODE * errors[2] = { 0, 0 }; /* [0] warnings, [1] errors */
static int    errcnt[2] = { 0, 0 };

/* the messages can go to max 3 output files/devices,
   default is just stderr */
#define MAXMSGFILES 3
static FILE * output[MAXMSGFILES] = { 0 /*stderr*/, 0, 0 };
static unsigned char setmsgfiles_called = 0;


void setmsgfiles( FILE * out1, FILE * out2, FILE * out3 )
/* ================================================================ */
{
   output[0] = out1;
   output[1] = out2;
   output[2] = out3;
   setmsgfiles_called = 1;
}


void geterrcnt( int * errors, int * warnings )
/* ================================================================ */
{
   *errors   = errcnt[MSG_ERROR  ];
   *warnings = errcnt[MSG_WARNING];
}


static void msg_out( FILE * f, int kind, SCAN * inp, int pos, char * txt, char * specific, char * srcfile, int srcline )
/* ---------------------------------------------------------------- */
/* put the message to 'f' */
{
   if( inp && (pos > 0) )
      fputs( prlineat( inp, pos), f );

   fprintf( f, " %s", kindstr[kind] );

   if( txt )
      fprintf( f, ": %s", txt );

   if( specific )
      fprintf( f, " (%s)", specific );

#ifndef _DEBUG
   if( kind == MSG_DEBUG )
#endif
      fprintf( f, " [%s,%d]", srcfile, srcline );

   fprintf( f, "\n" );
}


void msg( int kind, SCAN * inp, int pos, char * txt, char * specific, char * srcfile, int srcline )
/* ================================================================ */
/* a message to the user according to the kind constants below.
   inp/pos indicate error position, can be 0
   txt is the message, specific : some specific string like symbol name
   srcfile/srcline for __FILE__, __LINE__ */
{
char * s;
int newerr;
int i;

   if( !setmsgfiles_called )
      setmsgfiles( stderr, 0, 0 );     // default

   if( (kind < 0) || (kind > 3) )
   {
      fprintf( stderr, "unknown msg kind (%d) - assuming fatal error\n", kind );
      kind = MSG_FATAL;
   }

   if( (kind == MSG_ERROR) || (kind == MSG_WARNING) )
   {
      /* keep the essential information */
      s = strfalloc( "%s\001%d\001%s\001%d\001%s",
                     specific ? specific : " ",
                     pos,
                     inp ? inp->filename : " ",
                     srcline, srcfile );

      /* we need a tree: */
      if( !errors[kind] )
         errors[kind] = bt_mktree( "@", 0 );

      bt_insert( errors[kind], strlen(s), s, &newerr );

      if( !newerr )
      {
         /* not new */
         sfree( s );
         return; /* no need for printing it */
      }

      errcnt[kind]++; /* count them */
   }

   for( i = 0; i < MAXMSGFILES; i++ )
      if( output[i] )
         msg_out( output[i], kind, inp, pos, txt, specific, srcfile, srcline );

   if( kind == MSG_FATAL )
      exit( 3 );
/*
   if( kind == MSG_ERROR )
      exit( 4 );
*/
}


