/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* nstring.c
   routines for strings with explicite length
*/

#define min(A,B)  ((A)<=(B) ? A : B )

#include <string.h>
#include "nstring.h"



int nstrcmp( int s1len, char * s1, int s2len, char * s2 )
/* ================================================================ */
/* like strcmp but uses the given length.
   unsigned compare, null bytes are allowed */
{
int diff, i, minlen;
unsigned char * x1;
unsigned char * x2;

   x1 = (unsigned char *)s1;
   x2 = (unsigned char *)s2;
   minlen = min( s1len, s2len );
   for( i = 0; i < minlen; i++ )
   {
      diff = *x1++ - *x2++;
      if( diff )
         return diff;
   }
   return s1len - s2len;
}


int nstricmp( int s1len, char * s1, int s2len, char * s2 )
/* ================================================================ */
/* like stricmp but uses the given length.
   unsigned compare, null bytes are allowed */
{
int diff, i, minlen;
unsigned char * x1;
unsigned char * x2;

   x1 = (unsigned char *)s1;
   x2 = (unsigned char *)s2;
   minlen = min( s1len, s2len );
   for( i = 0; i < minlen; i++ )
   {
      diff = (*x1++ & ~0x20) - (*x2++ & ~0x20);
      if( diff )
         return diff;
   }
   return s1len - s2len;
}


int nstrcmp2z( int s1len, char * s1, char * s2 )
/* ================================================================ */
/* s2 zero-terminated */
{
   return nstrcmp( s1len, s1, strlen(s2), s2 );
}


int nstricmp2z( int s1len, char * s1, char * s2 )
/* ================================================================ */
/* s2 zero-terminated */
{
   return nstricmp( s1len, s1, strlen(s2), s2 );
}



#if 0 /* test */
#include <stdio.h>
void main( void )
/* ---------------------------------------------------------------- */
{

   printf( "%d (<0)\n", nstrcmp( 3, "abc", 3, "abd" ) );
   printf( "%d (<0)\n", nstrcmp( 3, "abc", 3, "ab}" ) );
   printf( "%d (=0)\n", nstrcmp( 2, "abc", 2, "abd" ) );
   printf( "%d (>0)\n", nstrcmp( 3, "abc", 2, "abd" ) );
   printf( "%d (<0)\n", nstrcmp( 3, "a\0006", 3, "a\0007" ) );
}
#endif
