/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* set.c
   set handling
   simple  set of int(0..CARDINAL-1)
*/

#include "set.h"


static IWORD mask[WORDSIZE] =
{
   0x00000001, 0x00000002, 0x00000004, 0x00000008,
   0x00000010, 0x00000020, 0x00000040, 0x00000080,
   0x00000100, 0x00000200, 0x00000400, 0x00000800,
   0x00001000, 0x00002000, 0x00004000, 0x00008000,
   0x00010000, 0x00020000, 0x00040000, 0x00080000,
   0x00100000, 0x00200000, 0x00400000, 0x00800000,
   0x01000000, 0x02000000, 0x04000000, 0x08000000,
   0x10000000, 0x20000000, 0x40000000, 0x80000000
};


int set_empty( _SET * s )
/* ================================================================ */
/* make an empty set
   returns 1 if ok, 0 if error (s invalid) */
{
int i;

   if( !s )
      return 0;

   for( i = 0; i < CARDINAL_WORDS; i++ )
      s->word[i] = 0;

   return 1;
}


int set_full( _SET * s )
/* ================================================================ */
/* make a full set - contains all elements of the base set
   returns 1 if ok, 0 if error (s invalid) */
{
int i;

   if( !s )
      return 0;

   for( i = 0; i < CARDINAL_WORDS; i++ )
      s->word[i] = (IWORD)(-1);

   return 1;
}


int set_put( _SET * s, unsigned int e )
/* ================================================================ */
/* put element e into set s
   returns 1 if ok, 0 if error (s or e invalid) */
{
int i;

   if( !s || (e >= CARDINAL) )
      return 0;

   i = e / WORDSIZE;
   s->word[i] |= mask[e & (WORDSIZE-1)];

   return 1;
}


int set_in( _SET * s, unsigned int e )
/* ================================================================ */
/* returns 1 if e in s, 0 otherwise (even for errors) */
{
int i;

   if( !s || (e >= CARDINAL) )
      return 0;

   i = e / WORDSIZE;
   return !!(s->word[i] & mask[e & (WORDSIZE-1)]);
}


int set_union( _SET * y, _SET * a, _SET * b )
/* ================================================================ */
/* y = a union b
   returns 1 if ok, 0 if error (one of the sets invalid) */
{
int i;

   if( !y || !a || !b )
      return 0;

   for( i = 0; i < CARDINAL_WORDS; i++ )
      y->word[i] = a->word[i] | b->word[i];

   return 1;
}


int set_intersect( _SET * y, _SET * a, _SET * b )
/* ================================================================ */
/* y = a intersect b
   returns 1 if ok, 0 if error (one of the sets invalid) */
{
int i;

   if( !y || !a || !b )
      return 0;

   for( i = 0; i < CARDINAL_WORDS; i++ )
      y->word[i] = a->word[i] & b->word[i];

   return 1;
}


int set_eq( _SET * a, _SET * b )
/* ================================================================ */
/* returns 1 if a == b, 0 otherwise (even in case of errors) */
{
int i;

   if( !a || !b )
      return 0;

   for( i = 0; i < CARDINAL_WORDS; i++ )
      if( a->word[i] != b->word[i] )
         return 0;

   return 1;
}


int set_power( _SET * s )
/* ================================================================ */
/* returns the number of elements in the given set,
   -1 if s is invalid */
{
int i, k, n;

   if( !s )
      return -1;

   n = 0;
   for( i = 0; i < CARDINAL_WORDS; i++ )
      for( k = 0; k < WORDSIZE; k++ )
         if( s->word[i] & mask[k] )
            n++;

   return n;
}


