#ifndef _SET_H
#define _SET_H
/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* set.h
   set handling
   simple  set of int(0..CARDINAL-1)
*/

#ifndef CARDINAL
#define CARDINAL 256
#endif

typedef unsigned int IWORD;
#define WORDSIZE (8*sizeof(IWORD))
#define CARDINAL_WORDS (((CARDINAL)+WORDSIZE-1)/WORDSIZE)


typedef struct
{
   IWORD word[CARDINAL_WORDS];
} _SET;

int set_empty( _SET * s );
/* make an empty set
   returns 1 if ok, 0 if error (s invalid) */

int set_full( _SET * s );
/* make a full set - contains all elements of the base set
   returns 1 if ok, 0 if error (s invalid) */

int set_put( _SET * s, unsigned int e );
/* put element e into set s
   returns 1 if ok, 0 if error (s or e invalid) */

int set_in( _SET * s, unsigned int e );
/* returns 1 if e in s, 0 otherwise (even for errors) */

int set_union( _SET * y, _SET * a, _SET * b );
/* y = a union b
   returns 1 if ok, 0 if error (one of the sets invalid) */

int set_intersect( _SET * y, _SET * a, _SET * b );
/* y = a intersect b
   returns 1 if ok, 0 if error (one of the sets invalid) */

int set_eq( _SET * a, _SET * b );
/* returns 1 if a == b, 0 otherwise (even in case of errors) */

int set_power( _SET * s );
/* returns the number of elements in the given set,
   -1 if s is invalid */

#endif /* _SET_H */

