/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/


#include <fcntl.h>
#include <malloc.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "port.h"
#include "salloc.h"
#include "sutil.h"

#include "mfatal.h" /* needed for MFATAL message */

#define  STRMAX  100000



char * itoa( int val, char * str, int radix )
/* ========================================== */
{      /* tricky: radix -10: unsigned decimal */
char * s;
char b[sizeof(int)*8+1];
int i;
char digit;
unsigned int uval;

   s = str;
   if( (val < 0) && (radix == 10) )
   {
      *s++ = '-';
      uval = -val;
   }
   else
      uval = val;

   if( radix < 0 )
      radix = -radix;

   i = 0;
   do
   {
      digit = uval % radix + '0';
      if( digit > '9' )
         digit += 7;
      b[i++] = digit;
   } while( (uval /= radix) > 0 );
   i--;
   for( ; i >= 0; i-- )
      *s++ = b[i];     /* reverse */

   *s = 0;
   return str;
}


char * strupr( char * s )
/* ================================================================ */
{
   char * t = s;
   while( *t )
   {
      if( (*t >= 'a') && (*t <= 'z') )
         *t += 'A' - 'a';

      t++;
   }
   return s;
}


char * strlwr( char * s )
/* ================================================================ */
{
   char * t = s;
   while( *t )
   {
      if( (*t >= 'A') && (*t <= 'Z') )
         *t += 'a' - 'A';

      t++;
   }
   return s;
}


char * strf( char * format, ... )
/* ================================================================ */
/* like printf - but
   to a static buffer which is returned */
{
static char buf[STRMAX];
va_list ap;
int bytes;

   va_start( ap, format );
   bytes = vsprintf( buf, format, ap );
   if( bytes >= STRMAX )
      MFATAL( "internal error: string too long", itoa( bytes, buf, 10 ) );
   va_end( ap );
   return buf;
}


char * strfalloc( char * format, ... )
/* ================================================================ */
/* like strf - but allocates a new buffer */
{
static char buf[STRMAX];
va_list ap;
char * newbuf;
int bytes;

   va_start( ap, format );
   bytes = vsprintf( buf, format, ap );
   if( bytes >= STRMAX )
      MFATAL( "internal error: string too long", itoa( bytes, buf, 10 ) );
   va_end( ap );
   newbuf = salloc( bytes + 1 );
   strcpy( newbuf, buf );
   return newbuf;
}


long filelength( int fh )
/* ================================================================ */
{
long current, ende;

   current = lseek( fh, 0L, SEEK_CUR );
   ende    = lseek( fh, 0L, SEEK_END );
   lseek( fh, current, SEEK_SET );
   return ende;
}


