// PL/0 interpreter
#include <stdio.h>
#include <stdlib.h>

#include "pl0scan.h"
#include "salloc.h"
#include "stack.h"


#define _QRINT_(X) /*printf X*/

#define TRUE  1
#define FALSE 0

typedef enum { kCONST, kVAR, kPROC } KIND;

typedef struct
{
   NSTRING name;
   KIND    kind;
   int     value; // if kPROC then pos of block
} ID;


STACK * g_id;

static void init( void );
int _program( SCAN * inp, int pos );


static ID * search( STACK * idstk, int namlen, char * namptr )
/* ---------------------------------------------------------------- */
/* im stack rueckwaerts suchen */
{
   ID * obj;
   int pos = stk_gettop(idstk);
   while( pos >= 0 )
   {
      obj = stk_get( idstk, pos );
      if( nstricmp( obj->name.len, obj->name.str, namlen, namptr ) == 0 )
         return obj;
      pos--;
   }
   return 0;
}


static int getvalue( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
/* const || var */
{
   ID * obj;
   int val;

   obj = search( g_id, strlenat(inp,pos), strptrat(inp,pos) );
   if( !obj )
   {
      MFATAL_POS( pos, "unknown identifier", prsymat(inp,pos) );
      return 1;
   }
   else if( (obj->kind == kCONST) || (obj->kind == kVAR) )
   {
      val = obj->value;
//      _QRINT_(( "(%.*s is %d)\n", nstrat(inp,pos), val ));
      return val;
   }
   else
   {
      MFATAL_POS( pos, "neither const nor variable", prsymat(inp,pos) );
      return 1;
   }
}


int main( int argc, char * argv[] )
/* ---------------------------------------------------------------- */
{
SCAN * inp;

   init();
   if( argc != 2 )
   {
      printf( "USAGE: %s pl0file\n", argv[0] );
      exit( 1 );
   }

   inp = screadfile( argv[1] );
   if( inp )
      inp = scan( inp );
   if( inp )
   {
      g_id = stk_create();
      _program( inp, nextsym(inp,0) );
   }
   else
      MFATAL( "scan errors in", argv[1] );

   exit( 0 );
}


//###################################################################
/* generated by xbnfgen 0.6
   input    = pl0.xbnf
   language = C

first[ assignment     ] = IDENT
first[ block          ] = CONST IDENT VAR PROCEDURE CALL BEGIN IF WHILE PRINT
first[ callstmt       ] = CALL
first[ condition      ] = '(' '+' '-' IDENT NUMBER ODD
first[ constblk       ] = CONST
first[ constdef       ] = IDENT
first[ expression     ] = '(' '+' '-' IDENT NUMBER
first[ factor         ] = '(' IDENT NUMBER
first[ ifstmt         ] = IF
first[ printexpr      ] = IDENT
first[ printstmt      ] = PRINT
first[ procdef        ] = PROCEDURE
first[ program        ] = CONST IDENT VAR PROCEDURE CALL BEGIN IF WHILE PRINT
first[ statement      ] = IDENT CALL BEGIN IF WHILE PRINT
first[ stmtblk        ] = BEGIN
first[ term           ] = '(' IDENT NUMBER
first[ varblk         ] = VAR
first[ vardef         ] = IDENT
first[ whilestmt      ] = WHILE

*/
#define INDENT 3
#define _PRINT_(X) /*printf X*/
static int level;

       int _program( SCAN * inp, int pos );
static int _block( SCAN * inp, int pos, int execute );
static int _constblk( SCAN * inp, int pos, int execute );
static int _constdef( SCAN * inp, int pos, int execute );
static int _varblk( SCAN * inp, int pos, int execute );
static int _vardef( SCAN * inp, int pos, int execute );
static int _procdef( SCAN * inp, int pos, int execute );
static int _statement( SCAN * inp, int pos, int execute );
static _SET first_1;
static int _assignment( SCAN * inp, int pos, int execute );
static int _callstmt( SCAN * inp, int pos, int execute );
static int _stmtblk( SCAN * inp, int pos, int execute );
static int _ifstmt( SCAN * inp, int pos, int execute );
static int _whilestmt( SCAN * inp, int pos, int execute );
static int _printstmt( SCAN * inp, int pos, int execute );
static int _printexpr( SCAN * inp, int pos, int execute );
static int _condition( SCAN * inp, int pos, int execute, int * pvalue );
static _SET first_2;
static _SET first_3;
static int _expression( SCAN * inp, int pos, int execute, int * pvalue );
static _SET first_4;
static int _term( SCAN * inp, int pos, int execute, int * pvalue );
static _SET first_5;
static int _factor( SCAN * inp, int pos, int execute, int * pvalue );
static _SET first_6;

static void init( void )
/* ---------------------------------------------------------------- */
{
   level = 0;

   set_empty( &first_1 );
   set_put( &first_1, IDENT );
   set_put( &first_1, CALL );
   set_put( &first_1, BEGIN );
   set_put( &first_1, IF );
   set_put( &first_1, WHILE );
   set_put( &first_1, PRINT );

   set_empty( &first_2 );
   set_put( &first_2, '(' );
   set_put( &first_2, '+' );
   set_put( &first_2, '-' );
   set_put( &first_2, IDENT );
   set_put( &first_2, NUMBER );
   set_put( &first_2, ODD );

   set_empty( &first_3 );
   set_put( &first_3, '#' );
   set_put( &first_3, '<' );
   set_put( &first_3, '=' );
   set_put( &first_3, '>' );
   set_put( &first_3, LEQ_ );
   set_put( &first_3, GEQ_ );

   set_empty( &first_4 );
   set_put( &first_4, '+' );
   set_put( &first_4, '-' );

   set_empty( &first_5 );
   set_put( &first_5, '*' );
   set_put( &first_5, '/' );

   set_empty( &first_6 );
   set_put( &first_6, '(' );
   set_put( &first_6, IDENT );
   set_put( &first_6, NUMBER );
}


 int _program( SCAN * inp, int pos )
/* ---------------------------------------------------------------- */
{
   _PRINT_(( "%*sprogram-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = _block( inp, pos, TRUE );
   pos = eatsym( inp, pos, '.' );
   level--;
   return pos;
}


static int _block( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   int tos, nerr, nwarn;

   _PRINT_(( "%*sblock-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   tos = stk_gettop(g_id);
   _QRINT_(( "block begin : %2d\n", tos ));
   if( symat(inp,pos) == CONST )
   {
      pos = _constblk( inp, pos, execute );
   }
   if( symat(inp,pos) == VAR )
   {
      pos = _varblk( inp, pos, execute );
   }
   while( symat(inp,pos) == PROCEDURE )
   {
      pos = _procdef( inp, pos, execute );
   }
   geterrcnt( &nerr, &nwarn );
   pos = _statement( inp, pos, execute && (nerr == 0) );
   stk_settop( g_id, tos );
   _QRINT_(( "block end  -> %2d\n", tos ));
   level--;
   return pos;
}


static int _constblk( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   _PRINT_(( "%*sconstblk-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = eatsym( inp, pos, CONST );
   pos = _constdef( inp, pos, execute );
   while( symat(inp,pos) == ',' )
   {
      pos = eatsym( inp, pos, ',' );
      pos = _constdef( inp, pos, execute );
   }
   pos = eatsym( inp, pos, ';' );
   level--;
   return pos;
}


static int _constdef( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   int tos, numpos;
   _NUMBER_ * pnum;
   ID * newobj;

   _PRINT_(( "%*sconstdef-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   if( execute )
   {
      newobj = salloc( sizeof(ID) );
      newobj->kind = kCONST;
      newobj->name.len = strlenat(inp,pos);
      newobj->name.str = strptrat(inp,pos);
   }
   pos = eatsym( inp, pos, IDENT );
   pos = eatsym( inp, pos, '=' );
   numpos = pos;
   pos = eatsym( inp, pos, NUMBER );
   pnum = NUMBERat(inp,numpos);
   if( execute )
   {
      newobj->value = pnum->intvalue;
      tos = stk_push( g_id, newobj );
      _QRINT_(( "%2d = const %.*s %d\n", tos, nstring(newobj->name), newobj->value ));
   }
   level--;
   return pos;
}


static int _varblk( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   _PRINT_(( "%*svarblk-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = eatsym( inp, pos, VAR );
   pos = _vardef( inp, pos, execute );
   while( symat(inp,pos) == ',' )
   {
      pos = eatsym( inp, pos, ',' );
      pos = _vardef( inp, pos, execute );
   }
   pos = eatsym( inp, pos, ';' );
   level--;
   return pos;
}


static int _vardef( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   int tos;
   ID * newobj;

   _PRINT_(( "%*svardef-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   if( execute )
   {
      newobj  = salloc( sizeof(ID) );
      newobj->kind = kVAR;
      newobj->value= 0;
      newobj->name.len = strlenat(inp,pos);
      newobj->name.str = strptrat(inp,pos);
   }
   pos = eatsym( inp, pos, IDENT );
   if( execute )
   {
      tos = stk_push( g_id, newobj );
      _QRINT_(( "%2d = var %.*s\n", tos, nstring(newobj->name) ));
   }
   level--;
   return pos;
}


static int _procdef( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   int tos;
   ID * newobj;

   _PRINT_(( "%*sprocdef-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = eatsym( inp, pos, PROCEDURE );
   if( execute )
   {
      newobj  = salloc( sizeof(ID) );
      newobj->kind = kPROC;
      newobj->name.len = strlenat(inp,pos);
      newobj->name.str = strptrat(inp,pos);
   }
   pos = eatsym( inp, pos, IDENT );
   pos = eatsym( inp, pos, ';' );
   if( execute )
      newobj->value = pos;
   pos = _block( inp, pos, FALSE );
   pos = eatsym( inp, pos, ';' );
   if( execute )
   {
      tos = stk_push( g_id, newobj );
      _QRINT_(( "%2d = procedure %.*s at %d\n", tos, nstring(newobj->name), newobj->value ));
   }
   level--;
   return pos;
}


static int _statement( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   _PRINT_(( "%*sstatement-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   if( set_in( &first_1, symat(inp,pos) ) )
   {
      switch( symat(inp,pos) )
      {
      case IDENT:
         pos = _assignment( inp, pos, execute );
         break;
      case CALL:
         pos = _callstmt( inp, pos, execute );
         break;
      case BEGIN:
         pos = _stmtblk( inp, pos, execute );
         break;
      case IF:
         pos = _ifstmt( inp, pos, execute );
         break;
      case WHILE:
         pos = _whilestmt( inp, pos, execute );
         break;
      case PRINT:
         pos = _printstmt( inp, pos, execute );
         break;
      default:
         MERROR( "expected one of", prtermsymset( inp, &first_1 ) );
         pos = nextsym( inp, pos );
      }
   }
   level--;
   return pos;
}


static int _assignment( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   int value;
   ID * var = 0;

   _PRINT_(( "%*sassignment-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   if( execute )
   {
      var = search( g_id, strlenat(inp,pos), strptrat(inp,pos) );
      if( !var )
      {
         MERROR( "unknown variable", prsymat(inp,pos) );
      }
      else if( var->kind != kVAR )
      {
         MERROR( "not a variable", prsymat(inp,pos) );
         var = 0;
      }
   }
   pos = eatsym( inp, pos, IDENT );
   pos = eatsym( inp, pos, ASSIGN_ );
   if( !var )
      execute = FALSE;
   pos = _expression( inp, pos, execute, &value );
   if( var )
   {
      _QRINT_(( "%.*s := %d\n", nstring(var->name), value ));
      var->value = value;
   }
   level--;
   return pos;
}


static int _callstmt( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   ID * proc;

   _PRINT_(( "%*scallstmt-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = eatsym( inp, pos, CALL );
   if( execute )
   {
      proc = search( g_id, strlenat(inp,pos), strptrat(inp,pos) );
      if( !proc )
      {
         MERROR( "unknown procedure", prsymat(inp,pos) );
      }
      else if( proc->kind != kPROC )
      {
         MERROR( "not a procedure", prsymat(inp,pos) );
      }
      else
      {
         _QRINT_(( "call %d\n", proc->value ));
         _block( inp, proc->value, TRUE ); // execute procedure block
      }
   }
   pos = eatsym( inp, pos, IDENT );
   level--;
   return pos;
}


static int _stmtblk( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   _PRINT_(( "%*sstmtblk-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = eatsym( inp, pos, BEGIN );
   pos = _statement( inp, pos, execute );
   while( symat(inp,pos) == ';' )
   {
      pos = eatsym( inp, pos, ';' );
      pos = _statement( inp, pos, execute );
   }
   pos = eatsym( inp, pos, END );
   level--;
   return pos;
}


static int _ifstmt( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   int cond;

   _PRINT_(( "%*sifstmt-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = eatsym( inp, pos, IF );
   pos = _condition( inp, pos, execute, &cond );
   pos = eatsym( inp, pos, THEN );
   pos = _statement( inp, pos, execute && cond );
   level--;
   return pos;
}


static int _whilestmt( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   int cond, condpos;

   _PRINT_(( "%*swhilestmt-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = eatsym( inp, pos, WHILE );
   condpos = pos;
 loop:
   pos = _condition( inp, pos, execute, &cond );
   pos = eatsym( inp, pos, DO );
   pos = _statement( inp, pos, execute && cond );
   if( execute && cond )
   {
      pos = condpos;
      goto loop;
   }
   level--;
   return pos;
}


static int _printstmt( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   _PRINT_(( "%*sprintstmt-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = eatsym( inp, pos, PRINT );
   if( execute )
   {
      _QRINT_(( "print :" ));
   }
   pos = _printexpr( inp, pos, execute );
   while( symat(inp,pos) == ',' )
   {
      pos = eatsym( inp, pos, ',' );
      pos = _printexpr( inp, pos, execute );
   }
   if( execute )
      printf( "\n" );
   level--;
   return pos;
}


static int _printexpr( SCAN * inp, int pos, int execute )
/* ---------------------------------------------------------------- */
{
   int varpos;

   _PRINT_(( "%*sprintexpr-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   varpos = pos;
   pos = eatsym( inp, pos, IDENT );
   if( symat(inp,pos) == '!' )
   {
      pos = eatsym( inp, pos, '!' );
      if( execute )
         printf( "  %.*s =", nstrat(inp,varpos) );
   }
   if( execute )
      printf( " %d", getvalue( inp, varpos ) );
   level--;
   return pos;
}


static int _condition( SCAN * inp, int pos, int execute, int * pvalue )
/* ---------------------------------------------------------------- */
{
   int val1, val2;
   _SYMBOL_ op;

   _PRINT_(( "%*scondition-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   switch( symat(inp,pos) )
   {
   case ODD:
      pos = eatsym( inp, pos, ODD );
      pos = _expression( inp, pos, execute, &val1 );
      *pvalue = val1 & 1;
      break;
   case '(':
   case '+':
   case '-':
   case IDENT:
   case NUMBER:
      pos = _expression( inp, pos, execute, &val1 );
      op = symat(inp,pos);
      switch( symat(inp,pos) )
      {
      case '=':
         pos = eatsym( inp, pos, '=' );
         break;
      case '#':
         pos = eatsym( inp, pos, '#' );
         break;
      case '<':
         pos = eatsym( inp, pos, '<' );
         break;
      case LEQ_:
         pos = eatsym( inp, pos, LEQ_ );
         break;
      case '>':
         pos = eatsym( inp, pos, '>' );
         break;
      case GEQ_:
         pos = eatsym( inp, pos, GEQ_ );
         break;
      default:
         MERROR( "expected one of", prtermsymset( inp, &first_3 ) );
         pos = nextsym( inp, pos );
      }
      pos = _expression( inp, pos, execute, &val2 );
      switch( op )
      {
      case '=':
         *pvalue = val1 == val2;
         break;
      case '#':
         *pvalue = val1 != val2;
         break;
      case '<':
         *pvalue = val1 <  val2;
         break;
      case LEQ_:
         *pvalue = val1 <= val2;
         break;
      case '>':
         *pvalue = val1 >  val2;
         break;
      case GEQ_:
         *pvalue = val1 >= val2;
         break;
      default:;
      }
      break;
   default:
      MERROR( "expected one of", prtermsymset( inp, &first_2 ) );
      pos = nextsym( inp, pos );
   }
   level--;
   return pos;
}


static int _expression( SCAN * inp, int pos, int execute, int * pvalue )
/* ---------------------------------------------------------------- */
{
   int val1, val2, sign;

   _PRINT_(( "%*sexpression-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   sign = 1;
   if( set_in( &first_4, symat(inp,pos) ) )
   {
      switch( symat(inp,pos) )
      {
      case '+':
         pos = eatsym( inp, pos, '+' );
         break;
      case '-':
         sign = -1;
         pos = eatsym( inp, pos, '-' );
         break;
      default:
         MERROR( "expected one of", prtermsymset( inp, &first_4 ) );
         pos = nextsym( inp, pos );
      }
   }
   pos = _term( inp, pos, execute, &val1 );
   val1 = val1 * sign;
   while( set_in( &first_4, symat(inp,pos) ) )
   {
      sign = 1;
      switch( symat(inp,pos) )
      {
      case '+':
         pos = eatsym( inp, pos, '+' );
         break;
      case '-':
         sign = -1;
         pos = eatsym( inp, pos, '-' );
         break;
      default:
         MERROR( "expected one of", prtermsymset( inp, &first_4 ) );
         pos = nextsym( inp, pos );
      }
      pos = _term( inp, pos, execute, &val2 );
      val1 = val1 + sign * val2;
   }
   *pvalue = val1;
   level--;
   return pos;
}


static int _term( SCAN * inp, int pos, int execute, int * pvalue )
/* ---------------------------------------------------------------- */
{
   int val1, val2, oppos;
   _SYMBOL_ op;

   _PRINT_(( "%*sterm-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   pos = _factor( inp, pos, execute, &val1 );
   while( set_in( &first_5, symat(inp,pos) ) )
   {
      oppos = pos;
      op = symat(inp,pos);
      switch( symat(inp,pos) )
      {
      case '*':
         pos = eatsym( inp, pos, '*' );
         break;
      case '/':
         pos = eatsym( inp, pos, '/' );
         break;
      default:
         MERROR( "expected one of", prtermsymset( inp, &first_5 ) );
         pos = nextsym( inp, pos );
      }
      pos = _factor( inp, pos, execute, &val2 );
      if( execute )
      {
         switch( op )
         {
         case '*':
            val1 = val1 * val2;
            break;
         case '/':
            if( val2 != 0 )
            {
               val1 = val1 / val2;
            }
            else
            {
               MFATAL_POS( oppos, "division by 0", "" );
            }
            break;
         default:;
         }
      }
   }
   *pvalue = val1;
   level--;
   return pos;
}


static int _factor( SCAN * inp, int pos, int execute, int * pvalue )
/* ---------------------------------------------------------------- */
{
   int numpos;

   _PRINT_(( "%*sfactor-(%s)\n", level*INDENT, "", prsymat(inp,pos) ));
   level++;
   switch( symat(inp,pos) )
   {
   case IDENT:
      if( execute )
         *pvalue = getvalue( inp, pos );
      pos = eatsym( inp, pos, IDENT );
      break;
   case NUMBER:
      numpos = pos;
      pos = eatsym( inp, pos, NUMBER );
      *pvalue = NUMBERat(inp,numpos)->intvalue;
      break;
   case '(':
      pos = eatsym( inp, pos, '(' );
      pos = _expression( inp, pos, execute, pvalue );
      pos = eatsym( inp, pos, ')' );
      break;
   default:
      MERROR( "expected one of", prtermsymset( inp, &first_6 ) );
      pos = nextsym( inp, pos );
   }
   level--;
   return pos;
}

