/* stack.c (fr) 2007
   simple stack
*/

#include <stdlib.h>
#include <string.h>

#include "salloc.h"
#include "stack.h"


/* Implementierung:
   Der Stack ist einfach ein dynamisch alloziertes Feld,
   wenn es nicht reicht, wird es verdoppelt...
   Der Inhalt ist einfach ein void Pointer. */

#define INITSIZE 1000



STACK * stk_create( void )
/* ================================================================ */
/* creates an empty stack and returns it via pstk,
   returns 1 if ok, 0 if error */
{
   STACK * pstk;


   pstk = salloc( sizeof(STACK) );

   pstk->parray = salloc( INITSIZE * sizeof(VOIDPTR) );
   pstk->size = INITSIZE;
   pstk->tos  = -1;      /* stack empty */
   return pstk;
}


int stk_push( STACK * pstk, void * val )
/* ================================================================ */
/* pushes val to the given stack, returns TOS (or < 0 for error) */
{
void * pnewarray;

   if( !pstk )
      return -1;

   pstk->tos++;
   if( pstk->tos >= pstk->size )
   {
      /* need a new array */
      pnewarray =  salloc( 2 * pstk->size * sizeof(void *) );
      memcpy( pnewarray, pstk->parray, pstk->size * sizeof(void *) );
      sfree( pstk->parray );
      pstk->parray = pnewarray;
      pstk->size *= 2;
   }
   /* put the value: */
   pstk->parray[pstk->tos] = val;
   return pstk->tos;
}


void * stk_pop( STACK * pstk )
/* ================================================================ */
/* pops the value at TOS and returns it (0 for errors) */
{
   if( !pstk || (pstk->tos < 0) )
      return 0;

   pstk->tos--;
   return pstk->parray[pstk->tos+1];
}


int stk_settop( STACK * pstk, int top )
/* ================================================================ */
/* if top < TOS -> sets TOS = top, otherwise no change, top >= -1
   returns TOS (< 0 for errors */
{
   int i;

   if( !pstk || (top < -1) )
      return -1;

   if( top < pstk->tos )
   {
      for( i = pstk->tos; i > top; i-- )
         sfree(pstk->parray[i]);

      pstk->tos = top;
   }

   return pstk->tos;
}


/* the remaining 2 functiona are implemented as macro in stack.h :
   stk_gettop(pstk), stk_get(pstk,index)  */

