/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* set2int.c
   maps sets to unique integers
*/

#include <malloc.h>
#include <memory.h>

#include "salloc.h"
#include "sutil.h"
#include "set2int.h"

#define NODEDATA int uniq;
#include "btree.h"


static NODE * sets = 0;
static int unumber = 0;


int set2int( _SET * set, int * isnew )
/* ================================================================ */
/* returns an integer which represents the given set:
   same number for equal sets, different number nor not equal sets
   isnew returns 1 if this is a new number */
{
NODE * thisset;
char * setalloc;

   if( !sets )
      sets = bt_mktree( "~", sizeof(NODE) );

   /* we just use the set as key in the binary tree .
      but a copy in the heap because the tree stores
      just pointers */
   setalloc = salloc( sizeof(_SET) );
   memcpy( setalloc, set, sizeof(_SET) );
   thisset = bt_insert( sets, sizeof(_SET), setalloc, isnew );
   if( *isnew )
   {
      unumber++;
      thisset->uniq = unumber;
   }
   else
      sfree(setalloc); /* don't need it twice */

   return thisset->uniq;
}


