/*
   xbnfgen version 0.6  Copyright (C) 2007-2014 Frank Reglin

   xbnfgen is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   xbnfgen is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
   See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with xbnfgen; see the file COPYING.
*/

/* xbnfscan.c
   xbnf scanner
   see xbnf.xbnf
   and scan.h
*/

//#define MAIN
// main is xbnf2ebnf !

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "xbnfscan.h"


#ifdef MAIN
#define LINEBREAK 9
char dump = 0;
#endif



SCAN * scan( SCAN * inp )
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
{
   int pos;
   char ch;

   scan_initauto( inp,
      0, 0,             // no comment block
      "#",              // start of comment until end of line
      "()[]{}<>.=|-",   // 1-char symbols
      0,                // no multi-char operators
      0,                // no keywords
      0, 0,             // no char literals
      TERMINAL, '"',    // strings are 'TERMINAL's
      0,                // feature flags
      0,                // no escape sequences
      0                 // no numbers
   );

   // scanning
   pos = 0;
   while( pos <= maxpos(inp) )
   {
      pos = scan_auto( inp, pos );
      if( charat(inp,pos) == '\\' )    // user symbol \n
      {
         pos = posinc( inp, pos, 1 );  /* \ */
         ch  = charat( inp, pos );
         if( ch != 'n' )
         {
            MERROR( "invalid char after backslash",
                 ch < ' ' ? strf("0x%02X",ch) : strf("'%c'=0x%02X",ch,ch) );
         }
#ifdef MAIN
         // we keep this as special symbol for EBNF output
         if( !dump )
            putsym( inp, pos, LINEBREAK );
#endif
         pos = posinc( inp, pos, 1 );  /* n */
      }
   }
   return inp;
}


int wordsymbol( SCAN * inp, int pos )
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
{
   char ch = charat(inp,pos);

   if( (ch >= 'a') && (ch <= 'z') )
      return LOWID;
   else
      return TERMINAL;
}


char * prusersymat( SCAN * inp, int pos )
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
{
   if( symat(inp,pos) == LOWID )
      return strf( "%.*s", nstrat(inp,pos) );
   else
      return "";
}


char * prusertermsym( int symbol )
/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */
{
   if( symbol == LOWID )
      return "identifier(nonterminal)";
   else
      return "";
}


#ifdef MAIN
// wir erzeugen hier EBNF als output (oder einen Dump)

void usage( void )
{
   printf( "USAGE: x2ebnf [-d] Inputfile\n" );
   printf( "        writes to stdout: -d : dump, otherwise : EBNF\n" );
   exit( 1 );
}


int main( int argc, char * argv[] )
/* ---------------------------------------------------------------- */
{
SCAN * inp;
int  pos;
char c;
char * input;

   if( (argc < 2) || (argc > 3) )
      usage();
   dump  = 0;
   input = argv[1];
   if( input[0] == '-' )
   {
      if( (input[1] == 'd') && (input[2] == 0) )
      {
         if( argc < 3 )
            usage();

         dump  = 1;
         input = argv[2];
      }
      else
         usage();
   }

   inp = screadfile( input );
   if( inp )
      inp = scan( inp );
   if( inp )
   {
      if( dump )
      {
         scan_dump( stdout, inp );
      }
      else
      {
         pos = 1;
         while( pos <= maxpos(inp) )
         {
            switch( symat(inp,pos) )
            {
            case NONE:
               break;
            case NXSYM:
               pos = attat(inp,pos).a - 1;
               break;
            case BOL:
               printf( "\n" );
               break;
            case LOWID:
               printf( " %.*s", nstrat(inp,pos) );
               break;
            case TERMINAL:
               c = charat(inp,pos); // erstes Zeichen
               if( ((c >= 'A') && (c <= 'Z')) || (c == '_') )
                  printf( " %.*s"    , textstringat(inp,pos) );
               else
                  printf( " \"%.*s\"", textstringat(inp,pos) );
               break;
            case EOFSYM:
               // printf( "end of file\n" );
               break;
            case LPAREN:  case RPAREN:
            case LBRACK:  case RBRACK:
            case LBRACE:  case RBRACE:
            case DOT:
            case EQ:
            case BAR:
               printf( " %c", symat(inp,pos) );
               break;
            case LT:
               printf( " (" );
               break;
            case GT:
               printf( " )" );
               break;
            case LINEBREAK:
               printf( " \\n" );
               break;
            case MINUS:
               printf(" { ANY }" );
               break;
            default:
               printf( " UNKNOWN{%.*s}\n", nstrat(inp,pos) );
               break;
            }
            pos++;
         } // while
      }
   }
   else
      MFATAL( "scan errors in", input );

   exit( 0 );
}
#endif
